;
; This file is part of GENFERM.
;
; Copyright 2005, 2022 Dario Alejandro Alpern
;
; GENFERM is free software: you can rrdistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation, either version 3 of the License, or
; (at your option) any later version.
;
; GENFERM is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with GENFERM.  If not, see <http://www.gnu.org/licenses/>.
;
_TEXT SEGMENT
; First argument pointed by RCX: First factor
; Second argument pointed by RDX: Second factor
; Third argument pointed by R8: Product
; Every element has 64 bits but only the lower 32 bits are used.
    PUBLIC MontgomeryMult
    EXTERN NumberLength:DWORD
    EXTERN Modulus:QWORD
    EXTERN MontgomeryMultN:DWORD

MontgomeryMult PROC
    push rbx        ; Save registers required by Visual Studio compiler.
    push rsi
    push rdi
    push rbp
    push r12
    push r13
    push r14
    push r15
    mov r9, rcx     ; Save pointer to first factor.
    mov r10, rdx    ; Save pointer to second factor.
    lea r13, Modulus
;    NumberLengthBytes = NumberLength * (int)sizeof(limb);
;    (void)memset(Prod, 0, NumberLengthBytes);
    cld
    mov rdi, r8     ; Point to product.
    mov rcx, 0      ; Erase it.
    mov ecx, NumberLength
    inc ecx
    mov eax, 0
    rep stosq
    mov ecx, NumberLength
    mov ebp, 0
;    for (int i = 0; i < NumberLength; i++)
;    {
    mov r11d, ecx  ; Use i backwards.
outer_loop:

;      int32_t Nbr = (factor1 + i)->x;
    mov r12d, dword ptr [r9]
    lea r9, [r9 + 8]

;      int64_t Pr = ((int64_t)Nbr * (int64_t)factor2->x) + (int64_t)Prod[0];
    mov eax, r12d          ; Get Nbr
    mul dword ptr [r10]    ; Multiply by factor2->x.
    add eax, [r8]          ; Add Prod[0].
    adc edx, 0
    mov esi, edx           ; ESI = High half of new value of Pr.

;      int32_t MontDig = ((int32_t)Pr * MontgomeryMultN[0].x) & MAX_INT_NBR;
    mul MontgomeryMultN
    mov r14d, eax          ; R14D <- MontDig

;      Pr -= (int64_t)MontDig * (int64_t)Modulus[0].x;
    mul dword ptr [r13]
    mov edi, 0
    sub esi, edx           ; ESI:EDI = New value of Pr.
    sbb ebx, ebx           ; Extend bits of Pr using carry flag.

;      for (int j = 1; j < NumberLength; j++)
;      {
    mov r15, 1
    jmp end_inner_loop
inner_loop:

;        Pr = ((int64_t)Nbr * (int64_t)(factor2 + j)->x) + (int64_t)Prod[j] +
;          (Pr >> BITS_PER_GROUP) -
;          ((int64_t)MontDig * (int64_t)Modulus[j].x);
    mov eax, r12d          ; Get Nbr.
    mul dword ptr [r10 + r15 * 8] ; Multiply by (factor2 + j)->x
    add eax, [r8 + r15 * 8]       ; Add Prod[j]
    adc edx, 0                    ; No carry can occur here.
    add eax, esi           ; Add Pr >> BITS_PER_GROUP
    adc edx, ebx           ; Use most significant bits of Pr.
    adc ebx, 0
    mov esi, edx           ; Save sum into ESI:EDI.
    mov edi, eax
    mov eax, r14d          ; Get MontDig.
    mul dword ptr [r13 + r15 * 8] ;Multiply by Modulus[j].x
    sub edi, eax           ; Subtract product.
    sbb esi, edx           ; ESI:EDI <- New value of Pr.
    sbb ebx, 0
;        Prod[j - 1] = (int)Pr & (int)MAX_INT_NBR_U;
    mov dword ptr [r8 + r15 * 8 - 8], edi
    inc r15d               ; j++
end_inner_loop:
    cmp r15d, ecx          ; if j < NumberLength...
    jb inner_loop          ; go back.
;      }
;      Prod[NumberLength - 1] = (int)(Pr >> BITS_PER_GROUP);
    add esi, ebp
    adc ebx, ebp
    mov ebp, ebx
    mov dword ptr [r8 + r15 * 8 - 8], esi ;Save into Prod[NumberLength - 1].
    dec r11d               ; Decrement i...
    jnz outer_loop         ; ...and jump if it is not zero.
;    }
;    if (Prod[NumberLength - 1] < 0)
;    {        // Prod < 0, so perform Prod <- Prod + Modulus.
    test ebx, ebx
    jns product_ok         ; Non negative means product OK.
;      unsigned int carry = 0U;
    clc
;      for (int idx = 0; idx < NumberLength; idx++)
;      {
    mov r15, 0
    mov r14d, ecx
;        carry = (unsigned int)Prod[idx] + (unsigned int)Modulus[idx].x +
;          (carry >> BITS_PER_GROUP);
;        Prod[idx] = carry & MAX_VALUE_LIMB;
adjust_product_loop:
    mov eax, [r13 + r15 * 8]   ; Get Modulus[idx].x
    adc [r8 + r15 * 8], eax    ; Add to Prod[idx]
    inc r15
    dec r14d
    jnz adjust_product_loop
;      }
product_ok:
;    }
    pop r15        ; Restore registers required by Visual Studio compiler.
    pop r14
    pop r13
    pop r12
    pop rbp
    pop rdi
    pop rsi
    pop rbx
    ret
MontgomeryMult ENDP
_TEXT ENDS
    end
