#
# This file is part of GENFERM.
#
# Copyright 2005, 2022 Dario Alejandro Alpern
#
# GENFERM is free software: you can rrdistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GENFERM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY# without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GENFERM.  If not, see <http://www.gnu.org/licenses/>.
#
    .text
# First argument pointed by RDI: First factor
# Second argument pointed by RSI: Second factor
# Third argument pointed by RDX: Product
# Every element has 64 bits but only the lower 32 bits are used.
    .global MontgomeryMult
    .extern NumberLength
    .extern Modulus
    .extern MontgomeryMultN

MontgomeryMult:
    push %rbx        # Save registers required by gcc compiler.
    push %rbp
    push %r12
    push %r13
    push %r14
    push %r15
    mov %rdx, %r8    # Save pointer to product.
    mov %rdi, %r9    # Save pointer to first factor.
    mov %rsi, %r10   # Save pointer to second factor.
    lea Modulus(%rip), %r13
#    NumberLengthBytes = NumberLength * (int)sizeof(limb)#
#    (void)memset(Prod, 0, NumberLengthBytes)#
    cld
    mov %r8, %rdi    # Point to product.
    mov $0, %rcx     # Erase it.
    mov NumberLength(%rip), %ecx
    inc %ecx
    mov $0, %eax
    rep stosq
    mov NumberLength(%rip), %ecx
    mov $0, %ebp
#    for (int i = 0# i < NumberLength# i++)
#    {
    mov %ecx, %r11d  # Use i backwards.
outer_loop:

#      int32_t Nbr = (factor1 + i)->x#
    mov (%r9), %r12d
    lea 8(%r9), %r9

#      int64_t Pr = ((int64_t)Nbr * (int64_t)factor2->x) + (int64_t)Prod[0]#
    mov %r12d, %eax         # Get Nbr
    mull (%r10)             # Multiply by factor2->x.
    add (%r8), %eax         # Add Prod[0].
    adc $0, %edx
    mov %edx, %esi          # ESI = High half of new value of Pr.

#      int32_t MontDig = ((int32_t)Pr * MontgomeryMultN) & MAX_INT_NBR#
    mull MontgomeryMultN(%rip)
    mov %eax, %r14d         # R14D <- MontDig

#      Pr -= (int64_t)MontDig * (int64_t)Modulus[0].x#
    mull (%r13)
    mov $0, %edi
    sub %edx, %esi          # ESI:EDI = New value of Pr.
    sbb %ebx, %ebx          # Extend bits of Pr using carry flag.

#      for (int j = 1# j < NumberLength# j++)
#      {
    mov $1, %r15
    jmp end_inner_loop
inner_loop:

#        Pr = ((int64_t)Nbr * (int64_t)(factor2 + j)->x) + (int64_t)Prod[j] +
#          (Pr >> BITS_PER_GROUP) -
#          ((int64_t)MontDig * (int64_t)Modulus[j].x)#
    mov %r12d, %eax          # Get Nbr.
    mull (%r10, %r15, 8)     # Multiply by (factor2 + j)->x
    add (%r8, %r15, 8), %eax # Add Prod[j]
    adc $0, %edx             # No carry can occur here.
    add %esi, %eax           # Add Pr >> BITS_PER_GROUP
    adc %ebx, %edx           # Use most significant bits of Pr.
    adc $0, %ebx
    mov %edx, %esi           # Save sum into ESI:EDI.
    mov %eax, %edi
    mov %r14d, %eax          # Get MontDig.
    mull (%r13, %r15, 8)     # Multiply by Modulus[j].x
    sub %eax, %edi           # Subtract product.
    sbb %edx, %esi           # ESI:EDI <- New value of Pr.
    sbb $0, %ebx
#        Prod[j - 1] = (int)Pr & (int)MAX_INT_NBR_U#
    mov %edi, -8(%r8, %r15, 8)
    inc %r15d                # j++
end_inner_loop:
    cmp %ecx, %r15d          # if j < NumberLength...
    jb inner_loop            # go back.
#      }
#      Prod[NumberLength - 1] = (int)(Pr >> BITS_PER_GROUP)#
    add %ebp, %esi
    adc %ebp, %ebx
    mov %ebx, %ebp
    mov %esi, -8(%r8, %r15, 8)# Save into Prod[NumberLength - 1].
    dec %r11d                # Decrement i...
    jnz outer_loop           # ...and jump if it is not zero.
#    }
#    if (Prod[NumberLength - 1] < 0)
#    {        // Prod < 0, so perform Prod <- Prod + Modulus.
    test %ebx, %ebx
    jns product_ok           # Non negative means product OK.
#      unsigned int carry = 0U#
    clc
#      for (int idx = 0# idx < NumberLength# idx++)
#      {
    mov $0, %r15
    mov %ecx, %r14d
#        carry = (unsigned int)Prod[idx] + (unsigned int)Modulus[idx].x +
#          (carry >> BITS_PER_GROUP)#
#        Prod[idx] = carry & MAX_VALUE_LIMB#
adjust_product_loop:
    mov (%r13, %r15, 8), %eax   # Get Modulus[idx].x
    adc %eax, (%r8, %r15, 8)    # Add to Prod[idx]
    inc %r15
    dec %r14d
    jnz adjust_product_loop
#      }
product_ok:
#    }
    pop %r15        # Restore registers required by gcc compiler.
    pop %r14
    pop %r13
    pop %r12
    pop %rbp
    pop %rbx
    ret
    